"""Basic tests for Powerfox."""

# pylint: disable=protected-access
import asyncio
from unittest.mock import patch

import pytest
from aiohttp import ClientError, ClientResponse, ClientSession
from aresponses import Response, ResponsesMockServer

from powerfox import Powerfox
from powerfox.exceptions import (
    PowerfoxConnectionError,
    PowerfoxError,
)

from . import load_fixtures


async def test_json_request(
    aresponses: ResponsesMockServer,
    powerfox_client: Powerfox,
) -> None:
    """Test JSON response is handled correctly."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/test",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixtures("all_devices.json"),
        ),
    )
    await powerfox_client._request("test")
    await powerfox_client.close()


async def test_internal_session(aresponses: ResponsesMockServer) -> None:
    """Test internal session is handled correctly."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/test",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixtures("all_devices.json"),
        ),
    )
    async with Powerfox(username="user", password="pass") as client:
        await client._request("test")


async def test_timeout(aresponses: ResponsesMockServer) -> None:
    """Test request timeout from the Powerfox API."""

    # Faking a timeout by sleeping
    async def response_handler(_: ClientResponse) -> Response:
        await asyncio.sleep(0.2)
        return aresponses.Response(
            body="Goodmorning!",
            text=load_fixtures("all_devices.json"),
        )

    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/test",
        "GET",
        response_handler,
    )

    async with ClientSession() as session:
        client = Powerfox(
            username="user",
            password="pass",
            session=session,
            request_timeout=0.1,
        )
        with pytest.raises(PowerfoxConnectionError):
            await client._request("test")


async def test_content_type(
    aresponses: ResponsesMockServer,
    powerfox_client: Powerfox,
) -> None:
    """Test content type is handled correctly."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/test",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "text/html"},
        ),
    )
    with pytest.raises(PowerfoxError):
        assert await powerfox_client._request("test")


async def test_client_error() -> None:
    """Test client error is handled correctly."""
    async with ClientSession() as session:
        client = Powerfox(username="user", password="pass", session=session)
        with (
            patch.object(
                session,
                "request",
                side_effect=ClientError,
            ),
            pytest.raises(PowerfoxConnectionError),
        ):
            assert await client._request("test")


async def test_response_status_404(
    aresponses: ResponsesMockServer,
    powerfox_client: Powerfox,
) -> None:
    """Test HTTP 404 response handling."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/test",
        "GET",
        aresponses.Response(status=404),
    )
    with pytest.raises(PowerfoxConnectionError):
        assert await powerfox_client._request("test")



"""Asynchronous Python client for Powerfox."""

from .exceptions import (
    PowerfoxAuthenticationError,
    PowerfoxConnectionError,
    PowerfoxError,
    PowerfoxNoDataError,
    PowerfoxPrivacyError,
    PowerfoxUnsupportedDeviceError,
)
from .local import PowerfoxLocal
from .models import (
    Device,
    DeviceReport,
    DeviceType,
    EnergyReport,
    GasReport,
    HeatMeter,
    LocalResponse,
    PowerMeter,
    Poweropti,
    ReportValue,
    WaterMeter,
)
from .powerfox import Powerfox

__all__ = [
    "Device",
    "DeviceReport",
    "DeviceType",
    "EnergyReport",
    "GasReport",
    "HeatMeter",
    "LocalResponse",
    "PowerMeter",
    "Powerfox",
    "PowerfoxAuthenticationError",
    "PowerfoxConnectionError",
    "PowerfoxError",
    "PowerfoxLocal",
    "PowerfoxNoDataError",
    "PowerfoxPrivacyError",
    "PowerfoxUnsupportedDeviceError",
    "Poweropti",
    "ReportValue",
    "WaterMeter",
]



"""Auth tests for PowerFox."""

# pylint: disable=protected-access
import pytest
from aresponses import ResponsesMockServer

from powerfox import Powerfox
from powerfox.exceptions import (
    PowerfoxAuthenticationError,
    PowerfoxError,
    PowerfoxNoDataError,
    PowerfoxPrivacyError,
    PowerfoxUnsupportedDeviceError,
)


async def test_authentication_error(
    aresponses: ResponsesMockServer,
    powerfox_client: Powerfox,
) -> None:
    """Test authentication error is handled correctly."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/test",
        "GET",
        aresponses.Response(status=401),
    )
    with pytest.raises(PowerfoxAuthenticationError):
        assert await powerfox_client._request("test")


async def test_no_poweropti_devices(
    aresponses: ResponsesMockServer,
    powerfox_client: Powerfox,
) -> None:
    """Test no Poweropti devices are found."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/my/all/devices",
        "GET",
        aresponses.Response(
            text="[]",
            headers={"Content-Type": "application/json"},
        ),
    )
    with pytest.raises(PowerfoxNoDataError):
        assert await powerfox_client.all_devices()


async def test_embedded_invalid_json_payload(
    aresponses: ResponsesMockServer,
    powerfox_client: Powerfox,
) -> None:
    """Test that an unparsable embedded payload is silently ignored."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/test",
        "GET",
        aresponses.Response(
            text='{"StatusCode": }',
            headers={"Content-Type": "application/json"},
        ),
    )
    # Malformed JSON -> decode raises ValueError -> guard returns, raw text returned
    result = await powerfox_client._request("test")
    assert '"StatusCode"' in result


async def test_embedded_low_status_code_payload(
    aresponses: ResponsesMockServer,
    powerfox_client: Powerfox,
) -> None:
    """Test that an embedded StatusCode below 400 is silently ignored."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/test",
        "GET",
        aresponses.Response(
            text='{"StatusCode": 200}',
            headers={"Content-Type": "application/json"},
        ),
    )
    # StatusCode < 400 -> guard returns, raw text returned
    result = await powerfox_client._request("test")
    assert result == '{"StatusCode": 200}'


async def test_embedded_generic_error_payload(
    aresponses: ResponsesMockServer,
    powerfox_client: Powerfox,
) -> None:
    """Test that an unknown embedded error status code raises PowerfoxError."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/my/test/current",
        "GET",
        aresponses.Response(
            text='{"StatusCode": 500, "ReasonPhrase": "Internal Server Error"}',
            headers={"Content-Type": "application/json"},
        ),
    )
    with pytest.raises(PowerfoxError):
        await powerfox_client.device("test")


async def test_embedded_precondition_failed_payload(
    aresponses: ResponsesMockServer,
    powerfox_client: Powerfox,
) -> None:
    """Test embedded API 412 in a HTTP 200 response is handled correctly."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/my/test/current",
        "GET",
        aresponses.Response(
            text=(
                '{"Version":"1.1","Content":{"Headers":[{"Key":"Content-Type",'
                '"Value":["text/plain; charset=utf-8"]}]},"StatusCode":412,'
                '"ReasonPhrase":"Precondition Failed","Headers":[],'
                '"TrailingHeaders":[]}'
            ),
            headers={"Content-Type": "application/json"},
        ),
    )
    with pytest.raises(PowerfoxPrivacyError):
        await powerfox_client.device("test")


@pytest.mark.parametrize("method_name", ["device", "raw_device_data", "report"])
async def test_no_poweropti_data(
    aresponses: ResponsesMockServer,
    powerfox_client: Powerfox,
    method_name: str,
) -> None:
    """Test no Poweropti data is found."""
    path = "/api/2.0/my/test/current"
    if method_name == "report":
        path = "/api/2.0/my/test/report"
    aresponses.add(
        "backend.powerfox.energy",
        path,
        "GET",
        aresponses.Response(
            text="{}",
            headers={"Content-Type": "application/json"},
        ),
    )
    method = getattr(powerfox_client, method_name)
    with pytest.raises(PowerfoxNoDataError):
        await method("test")


async def test_unsupported_poweropti_device(
    aresponses: ResponsesMockServer,
    powerfox_client: Powerfox,
) -> None:
    """Test unsupported Poweropti device error is raised."""
    # Simulate a FLOW response with Division: 4 (unsupported)
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/my/unsupported_device/current",
        "GET",
        aresponses.Response(
            text='{"Division": 4, "Timestamp": 1718812800, "Outdated": false}',
            headers={"Content-Type": "application/json"},
        ),
    )
    with pytest.raises(PowerfoxUnsupportedDeviceError) as exc:
        await powerfox_client.device("unsupported_device")
    assert "Division=4" in str(exc.value)



"""Fixtures for the Powerfox tests."""

from collections.abc import AsyncGenerator

import pytest
from aiohttp import ClientSession

from powerfox import Powerfox, PowerfoxLocal


@pytest.fixture(name="powerfox_client")
async def client() -> AsyncGenerator[Powerfox, None]:
    """Return a Powerfox client."""
    async with (
        ClientSession() as session,
        Powerfox(username="user", password="pass", session=session) as powerfox_client,
    ):
        yield powerfox_client


@pytest.fixture(name="powerfox_local_client")
async def local_client() -> AsyncGenerator[PowerfoxLocal, None]:
    """Return a PowerfoxLocal client."""
    async with (
        ClientSession() as session,
        PowerfoxLocal(
            host="192.168.1.50",
            api_key="1097bd725557",
            session=session,
        ) as powerfox_local_client,
    ):
        yield powerfox_local_client



"""Asynchronous Python client for Powerfox."""


class PowerfoxError(Exception):
    """Generic Powerfox exception."""


class PowerfoxConnectionError(PowerfoxError):
    """Powerfox connection exception."""


class PowerfoxAuthenticationError(PowerfoxError):
    """Powerfox authentication exception."""


class PowerfoxNoDataError(PowerfoxError):
    """Powerfox no data exception."""


class PowerfoxPrivacyError(PowerfoxError):
    """Powerfox privacy exception.

    Raised when the customer has denied data transmission (HTTP 412).
    """


class PowerfoxUnsupportedDeviceError(PowerfoxError):
    """Powerfox unsupported device exception."""



"""Example for the Powerfox report endpoint."""

import asyncio

from powerfox import Powerfox


async def main() -> None:
    """Show example on getting report data."""
    async with Powerfox(username="EMAIL_ADDRESS", password="PASSWORD") as client:
        report = await client.report(device_id="DEVICE_ID")
        print(report)

        # Example with date filters (year/month/day)
        report_filtered = await client.report(
            device_id="DEVICE_ID",
            year=2024,
            month=12,
            day=6,
        )
        print(report_filtered)


if __name__ == "__main__":
    asyncio.run(main())



"""Asynchronous Python client for Powerfox."""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import UTC, datetime
from enum import IntEnum
from typing import Any, ClassVar

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin


def _deserialize_timestamp(value: int | None) -> datetime | None:
    """Convert a timestamp to a datetime object."""
    if value is None:
        return None
    try:
        return datetime.fromtimestamp(value, tz=UTC)
    except (OverflowError, OSError, ValueError):
        return None


class DeviceType(IntEnum):
    """Enum for the different device types."""

    NO_TYPE = -1
    POWER_METER = 0
    COLD_WATER_METER = 1
    HOT_WATER_METER = 2
    HEAT_METER = 3
    GAS_METER = 4
    COLD_HOT_WATER_METER = 5

    @property
    def human_readable(self) -> str:
        """Return a human readable string for the device type."""
        return {
            DeviceType.POWER_METER: "Power Meter",
            DeviceType.COLD_WATER_METER: "Cold Water Meter",
            DeviceType.HOT_WATER_METER: "Hot Water Meter",
            DeviceType.HEAT_METER: "Heat Meter",
            DeviceType.GAS_METER: "Gas Meter",
            DeviceType.COLD_HOT_WATER_METER: "Cold/Hot Water Meter",
        }.get(self, "Unknown")


@dataclass
class Device(DataClassORJSONMixin):
    """Object representing a Device from Powerfox."""

    id: str = field(metadata=field_options(alias="DeviceId"))
    date_added: datetime = field(
        metadata=field_options(
            alias="AccountAssociatedSince",
            deserialize=lambda x: datetime.fromtimestamp(x, tz=UTC),
        )
    )
    main_device: bool = field(metadata=field_options(alias="MainDevice"))
    bidirectional: bool = field(metadata=field_options(alias="Prosumer"))
    type: DeviceType = field(metadata=field_options(alias="Division"))
    name: str = field(metadata=field_options(alias="Name"), default="Poweropti")


@dataclass
class Poweropti(DataClassORJSONMixin):
    """Object representing a Poweropti device."""

    outdated: bool = field(metadata=field_options(alias="Outdated"))
    timestamp: datetime = field(
        metadata=field_options(
            alias="Timestamp",
            deserialize=lambda x: datetime.fromtimestamp(x, tz=UTC),
        )
    )


@dataclass
class PowerMeter(Poweropti):
    """Object representing a Power device."""

    power: int = field(metadata=field_options(alias="Watt"))
    energy_usage: float | None = field(
        metadata=field_options(
            alias="A_Plus",
            deserialize=lambda x: x if x != 0 else None,
        ),
    )
    energy_return: float | None = field(
        metadata=field_options(
            alias="A_Minus",
            deserialize=lambda x: x if x != 0 else None,
        ),
    )
    energy_usage_high_tariff: float | None = field(
        metadata=field_options(alias="A_Plus_HT"), default=None
    )
    energy_usage_low_tariff: float | None = field(
        metadata=field_options(alias="A_Plus_NT"), default=None
    )


@dataclass
class HeatMeter(Poweropti):
    """Object representing a Heat device."""

    total_energy: int = field(metadata=field_options(alias="KiloWattHour"))
    delta_energy: int = field(metadata=field_options(alias="DeltaKiloWattHour"))
    total_volume: float = field(metadata=field_options(alias="CubicMeter"))
    delta_volume: float = field(metadata=field_options(alias="DeltaCubicMeter"))


@dataclass
class WaterMeter(Poweropti):
    """Object representing a Water device."""

    cold_water: float = field(metadata=field_options(alias="CubicMeterCold"))
    warm_water: float = field(metadata=field_options(alias="CubicMeterWarm"))


@dataclass
class ReportValue(DataClassORJSONMixin):
    """Object representing a report value entry."""

    device_id: str = field(metadata=field_options(alias="DeviceId"))
    timestamp: datetime = field(
        metadata=field_options(
            alias="Timestamp",
            deserialize=lambda x: datetime.fromtimestamp(x, tz=UTC),
        )
    )
    complete: bool = field(metadata=field_options(alias="Complete"))
    values_type: int | None = field(
        metadata=field_options(alias="ValuesType"),
        default=None,
    )
    delta: float | None = field(
        metadata=field_options(alias="Delta"),
        default=None,
    )
    delta_ht: float | None = field(
        metadata=field_options(alias="DeltaHT"),
        default=None,
    )
    delta_nt: float | None = field(
        metadata=field_options(alias="DeltaNT"),
        default=None,
    )
    delta_currency: float | None = field(
        metadata=field_options(alias="DeltaCurrency"),
        default=None,
    )
    total_delta: float | None = field(
        metadata=field_options(alias="TotalDelta"),
        default=None,
    )
    total_delta_currency: float | None = field(
        metadata=field_options(alias="TotalDeltaCurrency"),
        default=None,
    )
    consumption: float | None = field(
        metadata=field_options(alias="Consumption"),
        default=None,
    )
    consumption_kwh: float | None = field(
        metadata=field_options(alias="ConsumptionKWh"),
        default=None,
    )
    current_consumption: float | None = field(
        metadata=field_options(alias="CurrentConsumption"),
        default=None,
    )
    current_consumption_kwh: float | None = field(
        metadata=field_options(alias="CurrentConsumptionKwh"),
        default=None,
    )


@dataclass
class GasReport(DataClassORJSONMixin):
    """Object representing a gas report."""

    total_delta: float = field(metadata=field_options(alias="TotalDelta"))
    sum: float = field(metadata=field_options(alias="Sum"))
    total_delta_currency: float | None = field(
        metadata=field_options(alias="TotalDeltaCurrency"),
        default=None,
    )
    current_consumption_kwh: float | None = field(
        metadata=field_options(alias="CurrentConsumptionKwh"),
        default=None,
    )
    current_consumption: float | None = field(
        metadata=field_options(alias="CurrentConsumption"),
        default=None,
    )
    consumption_kwh: float | None = field(
        metadata=field_options(alias="ConsumptionKWh"),
        default=None,
    )
    consumption: float | None = field(
        metadata=field_options(alias="Consumption"),
        default=None,
    )
    max: float | None = field(
        metadata=field_options(alias="Max"),
        default=None,
    )
    max_currency: float | None = field(
        metadata=field_options(alias="MaxCurrency"),
        default=None,
    )
    max_consumption: float | None = field(
        metadata=field_options(alias="MaxConsumption"),
        default=None,
    )
    max_consumption_kwh: float | None = field(
        metadata=field_options(alias="MaxConsumptionKWh"),
        default=None,
    )
    min: float | None = field(
        metadata=field_options(alias="Min"),
        default=None,
    )
    min_consumption: float | None = field(
        metadata=field_options(alias="MinConsumption"),
        default=None,
    )
    min_consumption_kwh: float | None = field(
        metadata=field_options(alias="MinConsumptionKWh"),
        default=None,
    )
    avg_delta: float | None = field(
        metadata=field_options(alias="AvgDelta"),
        default=None,
    )
    avg_consumption: float | None = field(
        metadata=field_options(alias="AvgConsumption"),
        default=None,
    )
    avg_consumption_kwh: float | None = field(
        metadata=field_options(alias="AvgConsumptionKWh"),
        default=None,
    )
    meter_readings: list[dict[str, Any]] = field(
        default_factory=list,
        metadata=field_options(alias="MeterReadings"),
    )
    report_values: list[ReportValue] = field(
        default_factory=list,
        metadata=field_options(alias="ReportValues"),
    )
    sum_currency: float | None = field(
        metadata=field_options(alias="SumCurrency"),
        default=None,
    )


@dataclass
class EnergyReport(DataClassORJSONMixin):
    """Object representing an energy report section."""

    start_time: datetime | None = field(
        metadata=field_options(
            alias="StartTime",
            deserialize=_deserialize_timestamp,
        ),
        default=None,
    )
    start_time_currency: datetime | None = field(
        metadata=field_options(
            alias="StartTimeCurrency",
            deserialize=_deserialize_timestamp,
        ),
        default=None,
    )
    sum: float | None = field(
        metadata=field_options(alias="Sum"),
        default=None,
    )
    max: float | None = field(
        metadata=field_options(alias="Max"),
        default=None,
    )
    max_currency: float | None = field(
        metadata=field_options(alias="MaxCurrency"),
        default=None,
    )
    meter_readings: list[dict[str, Any]] = field(
        default_factory=list,
        metadata=field_options(alias="MeterReadings"),
    )
    report_values: list[ReportValue] = field(
        default_factory=list,
        metadata=field_options(alias="ReportValues"),
    )
    sum_currency: float | None = field(
        metadata=field_options(alias="SumCurrency"),
        default=None,
    )


@dataclass
class DeviceReport(DataClassORJSONMixin):
    """Object representing a report response."""

    gas: GasReport | None = field(
        default=None,
        metadata=field_options(alias="Gas"),
    )
    consumption: EnergyReport | None = field(
        default=None,
        metadata=field_options(alias="Consumption"),
    )
    feed_in: EnergyReport | None = field(
        default=None,
        metadata=field_options(alias="FeedIn"),
    )


@dataclass
class LocalResponse(DataClassORJSONMixin):
    """Object representing the local interface response.

    Maps OBIS codes from the local poweropti API to flat fields.
    Units: power in W, energy values in Wh.
    """

    _OBIS_MAP: ClassVar[dict[str, str]] = {
        "1.7.0": "power",
        "1.8.0": "energy_usage",
        "1.8.1": "energy_usage_high_tariff",
        "1.8.2": "energy_usage_low_tariff",
        "2.8.0": "energy_return",
    }

    timestamp: datetime = field(
        metadata=field_options(
            deserialize=lambda x: datetime.fromtimestamp(x, tz=UTC),
        )
    )
    power: int | None = None
    energy_usage: int | None = None
    energy_usage_high_tariff: int | None = None
    energy_usage_low_tariff: int | None = None
    energy_return: int | None = None

    @classmethod
    def __pre_deserialize__(cls, d: dict[str, Any]) -> dict[str, Any]:
        """Map OBIS values array to flat fields before deserialization."""
        for item in d.get("values", []):
            field_name = cls._OBIS_MAP.get(item["obis"])
            if field_name:
                d[field_name] = item["value"]
        return d



"""Asynchronous Python client for the Powerfox local interface."""

from __future__ import annotations

import asyncio
import socket
from dataclasses import dataclass
from importlib import metadata
from typing import Any, Self

from aiohttp import ClientError, ClientResponseError, ClientSession
from aiohttp.hdrs import METH_GET
from mashumaro.codecs.orjson import ORJSONDecoder
from yarl import URL

from .exceptions import (
    PowerfoxAuthenticationError,
    PowerfoxConnectionError,
    PowerfoxError,
)
from .models import LocalResponse

VERSION: str = metadata.version(__package__)  # ty:ignore[invalid-argument-type]


@dataclass
class PowerfoxLocal:
    """Client for the local poweropti REST API.

    Connects to a poweropti device on the local network via HTTP.
    Requires the powerfox PRO Service and firmware v2.02.07 or higher.
    """

    host: str
    api_key: str

    request_timeout: float = 10.0
    session: ClientSession | None = None

    _close_session: bool = False

    async def _request(
        self,
        uri: str,
        *,
        method: str = METH_GET,
    ) -> Any:
        """Handle a request to the local poweropti API.

        Args:
        ----
            uri: Request URI, for example, 'value'.
            method: HTTP method to use.

        Returns:
        -------
            The response text from the local poweropti API.

        Raises:
        ------
            PowerfoxConnectionError: An error occurred while communicating
                with the local poweropti API.
            PowerfoxAuthenticationError: Invalid or missing API key.
            PowerfoxError: Received an unexpected response from the API.

        """
        url = URL.build(
            scheme="http",
            host=self.host,
        ).join(URL(uri))

        headers = {
            "Accept": "application/json",
            "User-Agent": f"PythonPowerfox/{VERSION}",
            "X-API-KEY": self.api_key,
        }

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        try:
            async with asyncio.timeout(self.request_timeout):
                response = await self.session.request(
                    method,
                    url,
                    headers=headers,
                )
                response.raise_for_status()
        except TimeoutError as exception:
            msg = "Timeout occurred while connecting to local poweropti."
            raise PowerfoxConnectionError(msg) from exception
        except ClientResponseError as exception:
            if exception.status == 401:
                msg = "Authentication to local poweropti failed."
                raise PowerfoxAuthenticationError(msg) from exception
            msg = "Error occurred while communicating with local poweropti."
            raise PowerfoxConnectionError(msg) from exception
        except (ClientError, socket.gaierror) as exception:
            msg = "Error occurred while communicating with local poweropti."
            raise PowerfoxConnectionError(msg) from exception

        content_type = response.headers.get("Content-Type", "")
        if "application/json" not in content_type:
            text = await response.text()
            msg = "Unexpected content type response from local poweropti."
            raise PowerfoxError(
                msg,
                {"Content-Type": content_type, "Response": text},
            )

        return await response.text()

    async def value(self) -> LocalResponse:
        """Get current measurement data from the local poweropti.

        Returns
        -------
            The current measurement data.

        """
        response = await self._request("value")
        return ORJSONDecoder(LocalResponse).decode(response)

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> Self:
        """Async enter.

        Returns
        -------
            The PowerfoxLocal object.

        """
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.

        Args:
        ----
            _exc_info: Exec type.

        """
        await self.close()



"""Asynchronous Python client for Powerfox."""

import asyncio

from powerfox import Powerfox


async def main() -> None:
    """Show example on getting Powerfox data."""
    async with Powerfox(username="EMAIL_ADDRESS", password="PASSWORD") as client:
        poweropti = await client.device(device_id="DEVICE_ID")
        print(poweropti)


if __name__ == "__main__":
    asyncio.run(main())



"""Asynchronous Python client for the Powerfox local interface."""

import asyncio

from powerfox import PowerfoxLocal


async def main() -> None:
    """Show example on getting data from a local poweropti."""
    async with PowerfoxLocal(
        host="IP_ADDRESS",
        api_key="DEVICE_ID",
    ) as client:
        value = await client.value()
        print(value)


if __name__ == "__main__":
    asyncio.run(main())



"""Asynchronous Python client for Powerfox."""

from __future__ import annotations

import asyncio
import json
import socket
from dataclasses import dataclass
from importlib import metadata
from typing import Annotated, Any, Self

from aiohttp import BasicAuth, ClientError, ClientResponseError, ClientSession
from aiohttp.hdrs import METH_GET
from mashumaro.codecs.orjson import ORJSONDecoder
from mashumaro.exceptions import SuitableVariantNotFoundError
from mashumaro.types import Discriminator
from yarl import URL

from .exceptions import (
    PowerfoxAuthenticationError,
    PowerfoxConnectionError,
    PowerfoxError,
    PowerfoxNoDataError,
    PowerfoxPrivacyError,
    PowerfoxUnsupportedDeviceError,
)
from .models import Device, DeviceReport, Poweropti

VERSION: str = metadata.version(__package__)  # ty:ignore[invalid-argument-type]


@dataclass
class Powerfox:
    """Main class for handling connections with the Powerfox API."""

    username: str
    password: str

    request_timeout: float = 30.0
    session: ClientSession | None = None

    _close_session: bool = False

    def _raise_for_embedded_api_error(self, response_text: str) -> None:
        """Raise if Powerfox returned an API error envelope inside HTTP 200."""
        if '"StatusCode"' not in response_text:
            return

        try:
            api_response = json.loads(response_text)
        except json.JSONDecodeError:
            return

        if (
            not isinstance(api_response, dict)
            or not isinstance(status_code := api_response.get("StatusCode"), int)
            or status_code < 400
        ):
            return

        match status_code:
            case 412:
                msg = (
                    "Powerfox API refused data transfer, "
                    "the customer may have denied data transmission."
                )
                raise PowerfoxPrivacyError(msg)
            case _:
                msg = "Powerfox API returned an error in the response payload."
                raise PowerfoxError(
                    msg,
                    {
                        "StatusCode": status_code,
                        "ReasonPhrase": api_response.get(
                            "ReasonPhrase", "Unknown error"
                        ),
                        "Response": api_response,
                    },
                )

    async def _request(
        self,
        uri: str,
        *,
        method: str = METH_GET,
        params: dict[str, Any] | None = None,
    ) -> Any:
        """Handle a request to the Powerfox API.

        Args:
        ----
            uri: Request URI, without '/api/', for example, 'status'.
            method: HTTP method to use.
            params: Extra options to improve or limit the response.

        Returns:
        -------
            A Python dictionary (JSON decoded) with the response from
            the Powerfox API.

        Raises:
        ------
            PowerfoxConnectionError: An error occurred while communicating
                with the Powerfox API.
            PowerfoxError: Received an unexpected response from the Powerfox API.

        """
        url = URL.build(
            scheme="https",
            host="backend.powerfox.energy",
            path="/api/2.0/",
        ).join(URL(uri))

        headers = {
            "Accept": "application/json",
            "User-Agent": f"PythonPowerfox/{VERSION}",
        }

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        # Set basic auth credentials.
        auth = BasicAuth(self.username, self.password)

        try:
            async with asyncio.timeout(self.request_timeout):
                response = await self.session.request(
                    method,
                    url,
                    auth=auth,
                    headers=headers,
                    params=params,
                    ssl=True,
                )
                response.raise_for_status()
        except TimeoutError as exception:
            msg = "Timeout occurred while connecting to Powerfox API."
            raise PowerfoxConnectionError(msg) from exception
        except ClientResponseError as exception:
            if exception.status == 401:
                msg = "Authentication to the Powerfox API failed."
                raise PowerfoxAuthenticationError(msg) from exception
            msg = "Error occurred while communicating with Powerfox API."
            raise PowerfoxConnectionError(msg) from exception
        except (ClientError, socket.gaierror) as exception:
            msg = "Error occurred while communicating with Powerfox API."
            raise PowerfoxConnectionError(msg) from exception

        content_type = response.headers.get("Content-Type", "")
        if "application/json" not in content_type:
            text = await response.text()
            msg = "Unexpected content type response from Powerfox API."
            raise PowerfoxError(
                msg,
                {"Content-Type": content_type, "Response": text},
            )

        response_text = await response.text()
        self._raise_for_embedded_api_error(response_text)
        return response_text

    async def all_devices(self) -> list[Device]:
        """Get list of all Poweropti devices.

        Returns
        -------
            A list of all Poweropti devices.

        Raises
        ------
            PowerfoxNoDataError: If no devices are found or the response is empty.

        """
        response = await self._request("my/all/devices")
        if response == "[]":
            msg = "No Poweropti devices found."
            raise PowerfoxNoDataError(msg)
        return ORJSONDecoder(list[Device]).decode(response)

    async def device(self, device_id: str) -> Poweropti:
        """Get information about a specific Poweropti device.

        Args:
        ----
            device_id: The device ID to get information about.

        Returns:
        -------
            Information about the Poweropti device.

        Raises:
        ------
            PowerfoxNoDataError: If the response is empty or invalid JSON.

        """
        response = await self._request(
            f"my/{device_id}/current",
            params={"unit": "kwh"},
        )
        if response == "{}":
            msg = f"No data available for Poweropti device {device_id}."
            raise PowerfoxNoDataError(msg)

        try:
            return ORJSONDecoder(
                Annotated[Poweropti, Discriminator(include_subtypes=True)]
            ).decode(response)
        except SuitableVariantNotFoundError as err:
            data = ORJSONDecoder(dict).decode(response)
            division = data.get("Division", "unknown")
            msg = (
                "Unsupported device type received "
                f"(Division={division}) for device {device_id}."
            )
            raise PowerfoxUnsupportedDeviceError(msg) from err

    async def report(
        self,
        device_id: str,
        *,
        year: int | None = None,
        month: int | None = None,
        day: int | None = None,
    ) -> DeviceReport:
        """Get report information for a specific device.

        Args:
        ----
            device_id: The device ID to get report data for.
            year: Optional year to filter report data.
            month: Optional month to filter report data (requires year).
            day: Optional day to filter report data (requires year and month).

        Returns:
        -------
            Report data for the requested device.

        Raises:
        ------
            PowerfoxNoDataError: If the response is empty or invalid JSON.

        """
        if month is not None and year is None:
            msg = "Parameter 'month' requires 'year' to be set."
            raise ValueError(msg)
        if day is not None and (year is None or month is None):
            msg = "Parameter 'day' requires both 'year' and 'month' to be set."
            raise ValueError(msg)

        params: dict[str, Any] = {}
        if year is not None:
            params["year"] = year
        if month is not None:
            params["month"] = month
        if day is not None:
            params["day"] = day

        response = await self._request(
            f"my/{device_id}/report",
            params=params or None,
        )

        data = ORJSONDecoder(dict).decode(response)
        if not data:
            msg = f"No report data available for Poweropti device {device_id}."
            raise PowerfoxNoDataError(msg)

        return ORJSONDecoder(DeviceReport).decode(response)

    async def raw_device_data(self, device_id: str) -> dict[str, Any]:
        """Get raw JSON data for a specific Poweropti device.

        Args:
        ----
            device_id: The device ID to get information about.

        Returns:
        -------
            The raw response data as a dictionary.

        Raises:
        ------
            PowerfoxNoDataError: If the response is empty or invalid JSON.

        """
        response = await self._request(
            f"my/{device_id}/current",
            params={"unit": "kwh"},
        )

        data = ORJSONDecoder(dict).decode(response)
        if not data:
            msg = f"No data available for Poweropti device {device_id}."
            raise PowerfoxNoDataError(msg)
        return data

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> Self:
        """Async enter.

        Returns
        -------
            The Powerfox object.

        """
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.

        Args:
        ----
            _exc_info: Exec type.

        """
        await self.close()



"""Tests for the PowerfoxLocal client."""

# pylint: disable=protected-access
import asyncio
from unittest.mock import patch

import pytest
from aiohttp import ClientError, ClientResponse, ClientSession
from aresponses import Response, ResponsesMockServer
from syrupy.assertion import SnapshotAssertion

from powerfox import PowerfoxLocal
from powerfox.exceptions import (
    PowerfoxAuthenticationError,
    PowerfoxConnectionError,
    PowerfoxError,
)

from . import load_fixtures


async def test_value(
    aresponses: ResponsesMockServer,
    powerfox_local_client: PowerfoxLocal,
) -> None:
    """Test getting value from local poweropti."""
    aresponses.add(
        "192.168.1.50",
        "/value",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixtures("local_value.json"),
        ),
    )
    response = await powerfox_local_client.value()
    assert response.power == 228
    assert response.energy_usage == 17784955
    assert response.energy_usage_high_tariff == 17784955
    assert response.energy_usage_low_tariff == 0
    assert response.energy_return == 181


async def test_value_snapshot(
    aresponses: ResponsesMockServer,
    powerfox_local_client: PowerfoxLocal,
    snapshot: SnapshotAssertion,
) -> None:
    """Test value response matches snapshot."""
    aresponses.add(
        "192.168.1.50",
        "/value",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixtures("local_value.json"),
        ),
    )
    response = await powerfox_local_client.value()
    assert response == snapshot


async def test_internal_session(aresponses: ResponsesMockServer) -> None:
    """Test internal session is handled correctly."""
    aresponses.add(
        "192.168.1.50",
        "/value",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixtures("local_value.json"),
        ),
    )
    async with PowerfoxLocal(host="192.168.1.50", api_key="1097bd725557") as client:
        await client._request("value")


async def test_authentication_error(
    aresponses: ResponsesMockServer,
    powerfox_local_client: PowerfoxLocal,
) -> None:
    """Test authentication error handling."""
    aresponses.add(
        "192.168.1.50",
        "/value",
        "GET",
        aresponses.Response(status=401),
    )
    with pytest.raises(PowerfoxAuthenticationError):
        await powerfox_local_client._request("value")


async def test_timeout(aresponses: ResponsesMockServer) -> None:
    """Test request timeout from the local poweropti."""

    async def response_handler(_: ClientResponse) -> Response:
        await asyncio.sleep(0.2)
        return aresponses.Response(
            body="Goodmorning!",
            text=load_fixtures("local_value.json"),
        )

    aresponses.add(
        "192.168.1.50",
        "/value",
        "GET",
        response_handler,
    )

    async with ClientSession() as session:
        client = PowerfoxLocal(
            host="192.168.1.50",
            api_key="1097bd725557",
            session=session,
            request_timeout=0.1,
        )
        with pytest.raises(PowerfoxConnectionError):
            await client._request("value")


async def test_content_type(
    aresponses: ResponsesMockServer,
    powerfox_local_client: PowerfoxLocal,
) -> None:
    """Test content type is handled correctly."""
    aresponses.add(
        "192.168.1.50",
        "/value",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "text/html"},
        ),
    )
    with pytest.raises(PowerfoxError):
        await powerfox_local_client._request("value")


async def test_client_error() -> None:
    """Test client error is handled correctly."""
    async with ClientSession() as session:
        client = PowerfoxLocal(
            host="192.168.1.50",
            api_key="1097bd725557",
            session=session,
        )
        with (
            patch.object(
                session,
                "request",
                side_effect=ClientError,
            ),
            pytest.raises(PowerfoxConnectionError),
        ):
            await client._request("value")


async def test_response_status_404(
    aresponses: ResponsesMockServer,
    powerfox_local_client: PowerfoxLocal,
) -> None:
    """Test HTTP 404 response handling."""
    aresponses.add(
        "192.168.1.50",
        "/value",
        "GET",
        aresponses.Response(status=404),
    )
    with pytest.raises(PowerfoxConnectionError):
        await powerfox_local_client._request("value")



"""Asynchronous Python client for Powerfox."""

import asyncio

from powerfox import Powerfox


async def main() -> None:
    """Show example on getting Powerfox data."""
    async with Powerfox(username="EMAIL_ADDRESS", password="PASSWORD") as client:
        devices = await client.all_devices()
        print(devices)

        # How to show humand readable device type
        for device in devices:
            print(device.type.human_readable)


if __name__ == "__main__":
    asyncio.run(main())



"""Test the models for Powerfox."""

import pytest
from aresponses import ResponsesMockServer
from syrupy.assertion import SnapshotAssertion

from powerfox import (
    Device,
    DeviceReport,
    DeviceType,
    HeatMeter,
    LocalResponse,
    Powerfox,
    PowerMeter,
    Poweropti,
    WaterMeter,
)
from powerfox.models import _deserialize_timestamp

from . import load_fixtures


async def test_all_devices_data(
    aresponses: ResponsesMockServer,
    snapshot: SnapshotAssertion,
    powerfox_client: Powerfox,
) -> None:
    """Test devices data function."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/my/all/devices",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixtures("all_devices.json"),
        ),
    )
    devices: list[Device] = await powerfox_client.all_devices()
    assert devices == snapshot

    # Validate the human-readable property for each device
    for device in devices:
        assert isinstance(device.type, DeviceType)
        assert device.type.human_readable


async def test_power_meter_full_data(
    aresponses: ResponsesMockServer,
    snapshot: SnapshotAssertion,
    powerfox_client: Powerfox,
) -> None:
    """Test power meter full data function."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/my/power_device_id/current",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixtures("power_meter_full.json"),
        ),
    )
    power_meter: Poweropti = await powerfox_client.device("power_device_id")
    assert power_meter == snapshot
    assert isinstance(power_meter, PowerMeter)


async def test_power_meter_data(
    aresponses: ResponsesMockServer,
    snapshot: SnapshotAssertion,
    powerfox_client: Powerfox,
) -> None:
    """Test power meter data function."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/my/power_device_id/current",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixtures("power_meter.json"),
        ),
    )
    power_meter: Poweropti = await powerfox_client.device("power_device_id")
    assert power_meter == snapshot
    assert isinstance(power_meter, PowerMeter)
    assert power_meter.energy_usage_low_tariff is None
    assert power_meter.energy_usage_high_tariff is None


async def test_water_meter_data(
    aresponses: ResponsesMockServer,
    snapshot: SnapshotAssertion,
    powerfox_client: Powerfox,
) -> None:
    """Test water meter data function."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/my/water_meter_id/current",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixtures("water_meter.json"),
        ),
    )
    water_meter: Poweropti = await powerfox_client.device("water_meter_id")
    assert water_meter == snapshot
    assert isinstance(water_meter, WaterMeter)


async def test_heat_meter_data(
    aresponses: ResponsesMockServer,
    snapshot: SnapshotAssertion,
    powerfox_client: Powerfox,
) -> None:
    """Test heat meter data function."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/my/heat_meter_id/current",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixtures("heat_meter.json"),
        ),
    )
    heat_meter: Poweropti = await powerfox_client.device("heat_meter_id")
    assert heat_meter == snapshot
    assert isinstance(heat_meter, HeatMeter)


async def test_invalid_power_meter_data(
    aresponses: ResponsesMockServer,
    snapshot: SnapshotAssertion,
    powerfox_client: Powerfox,
) -> None:
    """Test invalid power meter data function."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/my/power_device_id/current",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixtures("power_meter_invalid.json"),
        ),
    )
    power_meter = await powerfox_client.device("power_device_id")
    assert power_meter == snapshot
    assert isinstance(power_meter, PowerMeter)
    assert power_meter.energy_usage is None
    assert power_meter.energy_return is None


async def test_raw_response_data(
    aresponses: ResponsesMockServer,
    snapshot: SnapshotAssertion,
    powerfox_client: Powerfox,
) -> None:
    """Test raw response data function."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/my/power_device_id/current",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixtures("power_meter.json"),
        ),
    )
    raw_data = await powerfox_client.raw_device_data("power_device_id")
    assert raw_data == snapshot


async def test_gas_report_data(
    aresponses: ResponsesMockServer,
    snapshot: SnapshotAssertion,
    powerfox_client: Powerfox,
) -> None:
    """Test gas report data function."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/my/flow_device_id/report",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixtures("gas_report.json"),
        ),
    )
    report: DeviceReport = await powerfox_client.report("flow_device_id")
    assert report == snapshot
    assert report.gas
    assert report.gas.report_values


async def test_power_report_data(
    aresponses: ResponsesMockServer,
    snapshot: SnapshotAssertion,
    powerfox_client: Powerfox,
) -> None:
    """Test powermeter report data function."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/my/power_device_id/report",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixtures("power_report.json"),
        ),
    )
    report: DeviceReport = await powerfox_client.report("power_device_id")
    assert report == snapshot
    assert report.consumption
    assert report.feed_in


async def test_report_with_filters(
    aresponses: ResponsesMockServer,
    powerfox_client: Powerfox,
) -> None:
    """Test report call with date filters."""
    aresponses.add(
        "backend.powerfox.energy",
        "/api/2.0/my/filter_device/report",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixtures("gas_report.json"),
        ),
    )
    report = await powerfox_client.report(
        "filter_device",
        year=2024,
        month=12,
        day=6,
    )
    assert report.gas


async def test_report_month_requires_year(powerfox_client: Powerfox) -> None:
    """Test passing month without year raises ValueError."""
    with pytest.raises(ValueError, match=r"month.*year"):
        await powerfox_client.report("filter_device", month=12)


async def test_report_day_requires_month(powerfox_client: Powerfox) -> None:
    """Test passing day without month/year raises ValueError."""
    with pytest.raises(ValueError, match=r"day.*year.*month"):
        await powerfox_client.report("filter_device", year=2024, day=6)


def test_deserialize_timestamp_none() -> None:
    """Ensure timestamp helper handles None values."""
    assert _deserialize_timestamp(None) is None


def test_local_response_unknown_obis() -> None:
    """Test that unknown OBIS codes are silently ignored."""
    response = LocalResponse.from_dict(
        {
            "timestamp": 1757053304,
            "values": [
                {"obis": "1.7.0", "value": 228},
                {"obis": "99.99.99", "value": 999},
            ],
        }
    )
    assert response.power == 228
    assert response.energy_usage is None
